#include <SPI.h>

/* For interacting with the SD card; SD.h was too big to fit in memory */
#include "PF.h"

/* For the gesture sensor */
#include <SparkFun_APDS9960.h>
#include "apds.h"
SparkFun_APDS9960 apds = SparkFun_APDS9960();
volatile int gesture_interrupt_flag;

/* For the Nokia 5110 display */
#include <Adafruit_PCD8544.h>
Adafruit_PCD8544 display = Adafruit_PCD8544(7, 6, 5, 8, 3);
#include "lcd_draw.h"

/* Number of options in main menu */
#define NUM_MENU_ITEMS 3
/* The currently selected main menu item */
int selected_menu_item;
/* For checking if we are in main menu or not */
volatile bool in_menu;

/* Buttons */
#define UP_BUTTON PC0
#define DOWN_BUTTON PC1
#define SELECT_BUTTON PC2

#define DEBOUNCE_TIME 250

volatile bool button_up_pressed;
volatile bool button_down_pressed;
volatile bool button_select_pressed;
volatile long lastPressed = 0;
long pressInterval = DEBOUNCE_TIME;

/* Leaderboard for the highest 10 scores */
#include "leaderboard.h"
Leaderboard leaderboard = Leaderboard();

/* Time for the first round of singleplayer/multiplayer modes */
#define START_TIME 11

#include "singleplayer.h"
Singleplayer singleplayer = Singleplayer(leaderboard);

#include "multiplayer.h"
Multiplayer multiplayer = Multiplayer();

/* Function for going back to the main menu after the end of a game */
void back_to_main_menu() {
  button_select_pressed = false;
  in_menu = true;

  TIMSK1 &= ~(1 << OCIE1A);

  draw_main_menu(selected_menu_item);
}

/* Interrupt for all the buttons; when pressed, a button's bool becomes true */
ISR(PCINT1_vect)
{
  if (millis() - lastPressed > pressInterval) {
    int ms = millis();
    lastPressed = millis();
    
    uint8_t pressed_mask = (~PINC);
    
    if (pressed_mask & (1 << UP_BUTTON)) {
      button_up_pressed = true;
    }
    
    if (pressed_mask & (1 << DOWN_BUTTON)) {
      button_down_pressed = true;
    }
    
    if (pressed_mask & (1 << SELECT_BUTTON)) {
      button_select_pressed = true;
    }
  }
}

void setup_buttons() {
  DDRC &= ~(1 << UP_BUTTON);
  PORTC |= (1 << UP_BUTTON);

  DDRC &= ~(1 << DOWN_BUTTON);
  PORTC |= (1 << DOWN_BUTTON);

  DDRC &= ~(1 << SELECT_BUTTON);
  PORTC |= (1 << SELECT_BUTTON);
  
  PCICR |= (1 << PCIE1);
  PCMSK1 |= (1 << PCINT8) | (1 << PCINT9) | (1 << PCINT10);
}

/* Interrupt that redirects to the instance's handler for the interrupt */
ISR(TIMER1_COMPA_vect) {
  if (singleplayer.started) {
    singleplayer.handleTimerInterrupt();
  } else if (multiplayer.started) {
    multiplayer.handleTimerInterrupt();
  }
}

void setup_timer1() {
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;

  // 1 Hz (16000000/((15624+1)*1024))
  OCR1A = 15624;
  // CTC
  TCCR1B |= (1 << WGM12);
  // Prescaler 1024
  TCCR1B |= (1 << CS12) | (1 << CS10);
}

void setup_apds() {
  pinMode(APDS9960_INT, INPUT);
  attachInterrupt(0, gestureInterrupt, FALLING);

  if (!apds.init()) {
    Serial.println(F("APDS9960 init failed"));
    while(1);
  }
}

/* Buzzer tones for different sound effects */
void buzzer_menu_change_sound() {
  tone(BUZZER_PIN, 500, 50);
}

void buzzer_select_back_sound() {
  tone(BUZZER_PIN, 800, 25);
}

void buzzer_correct_sound() {
  tone(BUZZER_PIN, 1000, 25);
  delay(25);
  tone(BUZZER_PIN, 1500, 25);
  delay(25);
}

void buzzer_incorrect_sound() {
  tone(BUZZER_PIN, 200, 50);
}

void buzzer_game_over_sound() {
  tone(BUZZER_PIN, 100, 200);
  delay(300);
  tone(BUZZER_PIN, 50, 400);
}

void buzzer_new_game_sound() {
  tone(BUZZER_PIN, 1000, 50);
  delay(50);
  tone(BUZZER_PIN, 1200, 50);
  delay(50);
  tone(BUZZER_PIN, 1400, 50);
  delay(50);
}

void setup() {
  Serial.begin(9600);

  if (disk_initialize()) {
    Serial.println(F("Error: disk_initialize"));
    while(1);
  }

  randomSeed(analogRead(0));

  setup_apds();
  pinMode(BUZZER_PIN, OUTPUT);
  setup_display();
  setup_buttons();
  setup_timer1();
  
  in_menu = true;
  draw_main_menu(selected_menu_item);

  sei();
}

void loop() {
  /* If in main menu, change what is displayed if button up or down are pressed
   * Otherwise, check which option was selected and go to the specific code (singleplayer, multiplayer, leaderboard)
   */
  if (in_menu) {
    bool menu_changed = false;
    
    if (button_up_pressed) {
      if (selected_menu_item > 0) {
        menu_changed = true;
        selected_menu_item--;

        buzzer_menu_change_sound();
      }
      
      button_up_pressed = false;
    } else if (button_down_pressed) {
      if (selected_menu_item < NUM_MENU_ITEMS - 1) {
        menu_changed = true;
        selected_menu_item++;
        buzzer_menu_change_sound();
      }
      
      button_down_pressed = false;
    } else if (button_select_pressed) {
      button_select_pressed = false;
      in_menu = false;

      buzzer_select_back_sound();
    }
 
    if (menu_changed)
      draw_main_menu(selected_menu_item);
  } else {
    if (selected_menu_item == 0) {
      singleplayer.start(START_TIME, 3);
    } else if (selected_menu_item == 1) {
      multiplayer.start(START_TIME, 3);
    } else if (selected_menu_item == 2) {
      leaderboard.getScores();
      draw_leaderboard(leaderboard);

      /* Stay in leaderboard until the back (select) button is pressed */
      while (true) {
        if (button_select_pressed) {
          leaderboard.resetAttributes();
          buzzer_select_back_sound();
          back_to_main_menu();
          break;
        }

        if (button_up_pressed) {
          leaderboard.pageUp();
          button_up_pressed = false; 
        } else if (button_down_pressed) {
          leaderboard.pageDown();
          button_down_pressed = false;
        } 
      }
    }
  }
}
